::------------------------------------------------------------------------------
:: NAME
::     kbc.bat - Keal's Batch Compiler
::
:: DESCRIPTION
::     A wrapper for FBC32 and HTML generator for error files
::
:: ARGUMENTS
::     %1 - The full path to the folder containing .bas files
::
:: AUTHOR
::     sintrode
::     NeinstarFilmz
::------------------------------------------------------------------------------
@echo off
setlocal enabledelayedexpansion

call :validateArguments "%~1" || goto :usage
call :setDirectories "%~f1" "%~f0"

call :generateDirectories
call :initializeLogFiles

call :generateIconFile
call :generateDebuggerScript

call :startHTMLReport
call :compileBASFile
call :parseCompilerLog
call :completeHTMLReport
exit /b

::------------------------------------------------------------------------------
:: Check to see if a valid path was passed to the script as an argument
::
:: Arguments: %1 - The argument passed to the script, if any
:: Returns:   0 if a valid path was passed to the script
::            1 otherwise
::------------------------------------------------------------------------------
:validateArguments
set "is_valid=1"
if exist "%~f1" set "is_valid=0"
exit /b %is_valid%

::------------------------------------------------------------------------------
:: Sets values for input and output directories
::
:: Arguments: %1 - The path of the .bas folder that was passed to the script
::            %2 - The full path of the main script
:: Returns:   None
::------------------------------------------------------------------------------
:setDirectories
set "project_path=%~dp1"
:: Strip the trailing backslash so that we can pretend it's a filename
for %%A in ("!project_path:~0,-1!") do set "project_basename=%%~nA"

for %%A in ("%~dp2..") do set "devkit_root=%%~fA"
set "output_root=%~dp2"
set "output_root=!output_root:~0,-1!" %= for purely aesthetic reasons =%
set "output_dir=!output_root!\!project_basename!"
set "output_temp_dir=!output_dir!\temp"
set "output_asset_dir=!output_dir!\assets"
exit /b

::------------------------------------------------------------------------------
:: Creates folders if they do not already exist
::
:: Arguments: None
:: Returns:   None
::------------------------------------------------------------------------------
:generateDirectories
for %%A in ("!output_dir!" "!output_temp_dir!" "!output_asset_dir!") do (
    if not exist "%%~A" mkdir "%%~A"
)
exit /b

::------------------------------------------------------------------------------
:: Clears any previous maintenance records or compiler logs
::
:: Arguments: None
:: Returns:   None
::------------------------------------------------------------------------------
:initializeLogFiles
set "logfile_maintenance=!output_temp_dir!\!project_basename! (Maintenance).txt"
set "logfile_compile=!project_path!\!project_basename! (Compile Log).html"
set "logfile_raw=!output_temp_dir!\!project_basename! (Raw Compiler Messages).txt"
set "logfile_header=!output_temp_dir!\!project_basename! (Temp Header).html"

for /f "tokens=2 delims==" %%A in ('set logfile_') do (
    if exist "%%~A" del /f "%%~A"
    >"%%~A" type nul
    >>"!logfile_maintenance!" echo Wipe Direct: "%%~A"
)
exit /b

::------------------------------------------------------------------------------
:: Copies a blank icon file into place and creates an associated .rc file
::
:: Arguments: None
:: Returns:   None
::------------------------------------------------------------------------------
:generateIconFile
set "icon_path=!output_asset_dir!\!project_basename!.ico"
set "blank_icon_path=!output_root!\blank disk.ico"

if not exist "!icon_path!" (
    if exist "!blank_icon_path!" (
        copy "!blank_icon_path!" "!icon_path!" >nul
    )
)
>"!output_asset_dir!\!project_basename!.rc" echo FB_PROGRAM_ICON ICON "!icon_path!"
exit /b

::------------------------------------------------------------------------------
:: Creates a script for passing the generated .exe file to fbdebug
::
:: Arguments: None
:: Returns:   None
::------------------------------------------------------------------------------
:generateDebuggerScript
>"!output_dir!\!project_basename! (SARG's FB Debugger).cmd" (
    echo echo SARG's FB Debugger
    echo "!output_root!\fbdebug_WDS_LNX\Windows 32bit\fbdebugger_new.exe" "!project_path!\!project_basename!.exe"
)
exit /b

::------------------------------------------------------------------------------
:: Creates the first half of the output HTML file. We're splitting the log
:: generation into two parts because the compiler output goes in its own file
:: and we're processing it later.
::
:: Arguments: None
:: Returns:   None
::------------------------------------------------------------------------------
:startHTMLReport
set "short_exe_path=!project_basename!.exe"
set "full_exe_path=!project_path!!short_exe_path!"

>"!logfile_header!" (
    REM == HEADER ==
    echo ^<^^!DOCTYPE html^>
    echo ^<html lang="en-us"^>
    echo ^<head^>
    echo ^<meta charset="UTF-8"^>
    echo ^<meta name="Author" content="trkeal@gmail.com http://puzzlum.net"^>
    echo ^<meta name="Classification" content="project compilation"^>
    echo ^<meta name="Description" content="!project_basename! (Keal's Batch Compiler log)"^>
    echo ^<meta name="Keywords" content="!project_basename!"^>
    echo ^<title^>"!project_basename!" Compiler Log ^(in HTML^)^</title^>
    
    for /f "delims=" %%A in ('type "!output_root!\favicon.b64"') do (
        echo ^<link rel="icon" type="image/png" href="%%~A"^>
    )
    
    echo ^<style^>
    type "!output_root!\stylesheet.css"
    echo ^</style^>

    echo ^</head^>
    echo ^<body^>
    echo ^<table^>

    REM == BANNER ==
    set /a "OddEven=0"
    call :alternator

    for /f "delims=" %%A in ('type "!output_root!\favicon.b64"') do (
        echo ^<img style="height: 4.8em; float: left; width: auto; padding: 0.2em;" src="%%~A" alt="Blank Diskette"^>
    )

    echo ^<hr^>
    echo ^<span class="header"^>"!project_basename!"^</span^>
    echo ^<hr^>
    echo ^<span class="header"^>Compiler Log ^(in HTML^)^</span^>
    echo ^<hr^>
    echo ^<span class="header2"^>%date% %time%^</span^>
    echo ^<hr^>
    echo ^</td^>^</tr^>

    REM == SUMMARY ==
    if exist "!project_path!\!project_basename!.exe" (
        call :alternator
        echo ^<hr^>
        echo ^<span^>Now Pruning: "!full_exe_path!"^</span^>
        echo ^</td^>^</tr^>

        call :alternator
        echo ^<hr^>
        echo ^<span^>Now Force Quitting: "!short_exe_path!"^</span^>
        echo ^<hr^>
        echo ^</td^>^</tr^>

        call :killProcess

        call :alternator
        echo ^<span^>Now Deleting Binary: "!full_exe_path!"^</span^>
        echo ^<hr^>
        echo ^</td^>^</tr^>
        if exist "!full_exe_path!" del "!full_exe_path!"
        if exist "!full_exe_path!" (
            call :alternator
            echo ^<span^>ERROR: Failed to delete the binary "!full_exe_path!"^</span^>
            echo ^<hr^>
            echo ^<span^>Must Prune Before Compiling^</span^>
            echo ^</td^>^</tr^>

            call :endHTMLReport
            exit /b
        )
    )
)
exit /b

::------------------------------------------------------------------------------
:: Compiles the .bas file and stores the compiler output in a log file
::
:: Arguments: None
:: Returns:   None
::------------------------------------------------------------------------------
:compileBASFile
set "options_compiler_exe="!devkit_root!\FreeBASIC\fbc32.exe""
set "options_compiler_output=-x "!project_basename!.exe""
set "options_compiler_bas="!project_path!!project_basename!.bas""
set "options_compiler_build=-asm intel"
set "options_compiler_gui=-s gui "!output_asset_dir!\!project_basename!.rc""
set "options_compiler_debug=-exx -g -gen gas"
set "options_compiler_lib="
set "options_compiler_module=-rr -m "!project_basename!" !options_compiler_bas!"

set compiler_string=!options_compiler_exe! !options_compiler_build! !options_compiler_debug! !options_compiler_lib! !options_compiler_module!
!compiler_string! >"!logfile_raw!" 2>&1

if "!errorlevel!"=="0" (
    for %%A in (asm o exe) do set "output_file_%%A=!project_path!\!project_basename!.%%A"
    for %%A in (obj rc) do set "output_file_%%A=!output_asset_dir!\!project_basename!.%%A"
    
    set "options_assembler_path=!devkit_root!\FreeBASIC\bin\win32\as.exe"
    set "options_rc_path=!devkit_root!\FreeBASIC\bin\win32\GoRC.exe"
    set "options_linker_path=!devkit_root!\FreeBASIC\bin\win32\ld.exe"
    set "options_lib_dir=!devkit_root!\FreeBASIC\lib\win32"

    set "options_assembler_args=--32 "!output_file_asm!" -o "!output_file_o!""
    set "options_rc_args=/ni /nw /o /fo "!output_file_obj!" "!output_file_rc!""
    set "options_linker_args=-m i386pe -o "!output_file_exe!" -subsystem windows -T "!options_lib_dir!\fbextra.x" --stack 1048576,1048756 -L "!options_lib_dir!" -L ".""
    set "options_linker_args=!options_linker_args! "!options_lib_dir!\crt2.o" "!options_lib_dir!\crtbegin.o" "!options_lib_dir!\fbrt0.o" "!output_file_o!" "!output_file_obj!""
    set "options_linker_args=!options_linker_args! "-^(" -lfb -lfbgfx -lgdi32 -lwinmm -lgcc -lmsvcrt -lkernel32 -luser32 -lmingw32 -lmingwex -lmoldname -lgcc_eh "-^)""
    set "options_linker_args=!options_linker_args! "!options_lib_dir!\crtend.o""

    >>"!logfile_raw!" 2>&1 (
        "!options_assembler_path!" !options_assembler_args!
        "!options_rc_path!" !options_rc_args!
        "!options_linker_path!" !options_linker_args!
    )
    del /q "!output_file_asm!" "!output_file_o!" "!output_file_obj!" 2>nul
)

>>"!logfile_header!" (
    call :alternator

	echo ^<span style="color: red;"^>Compiling Using Following Options:^</span^>
	echo ^<ul^>
	echo ^<li^>^<span class="caret"^>Binary:^</span^>^<ul class="nested"^>^<li^>^<span class="caret"^>!options_compiler_exe!^</span^>^</li^>^</ul^>^</li^>
	echo ^<li^>^<span class="caret"^>Source:^</span^>^<ul class="nested"^>^<li^>^<span class="caret"^>!options_compiler_bas!^</span^>^</li^>^</ul^>^</li^>
	echo ^<li^>^<span class="caret"^>Build: ^</span^>^<ul class="nested"^>^<li^>^<span class="caret"^>!options_compiler_build!^</span^>^</li^>^</ul^>^</li^>
	echo ^<li^>^<span class="caret"^>Gui:   ^</span^>^<ul class="nested"^>^<li^>^<span class="caret"^>!options_compiler_gui!^</span^>^</li^>^</ul^>^</li^>
	echo ^<li^>^<span class="caret"^>Debug: ^</span^>^<ul class="nested"^>^<li^>^<span class="caret"^>!options_compiler_debug!^</span^>^</li^>^</ul^>^</li^>
	echo ^<li^>^<span class="caret"^>Lib:   ^</span^>^<ul class="nested"^>^<li^>^<span class="caret"^>!options_compiler_lib!^</span^>^</li^>^</ul^>^</li^>
	echo ^</ul^>
    echo ^<button onclick="to_clipboard()" id="copybutton"^>Selected Options ^(0^)^</button^>
    echo ^<code id="clip"^>^</code^>
	echo ^</td^>^</tr^>
)
exit /b

::------------------------------------------------------------------------------
:: Extract compiler errors and linker errors from the compiler log
::
:: Arguments: None
:: Returns:   None
::------------------------------------------------------------------------------
:parseCompilerLog
set "err_count=0"
set "warn_count=0"
set "msg_count=0"

for %%A in ("error,err_count" "warning,warn_count" ":,msg_count") do (
    for /f "tokens=1,2 delims=," %%B in ("%%~A") do (
        for /f "tokens=3 delims=:" %%D in ('find /c "%%B" "!logfile_raw!"') do (
            set /a "%%C=%%D"
        )
    )
)

set "errwarn_index=0"
for /f "usebackq tokens=1,2,* delims=:" %%A in ("!logfile_raw!") do (
    set "error_line=%%A:%%B"

    if "!error_line!"=="!options_linker_path!" (
        for /f "tokens=1-3,* delims=:" %%D in ("%%~C") do (
            if not "%%G"=="" (
                set /a errwarn_index+=1
                set "errwarn[!errwarn_index!][filename]=%%~D:%%~E"
                set "errwarn[!errwarn_index!][line_num]=%%~F"
                set "errwarn[!errwarn_index!][type]=linker"
                set "errwarn[!errwarn_index!][error_msg]=%%~G"
                set "errwarn[!errwarn_index!][error_code]=N/A"
            )
        )
    ) else (
        set /a errwarn_index+=1
        set "errwarn[!errwarn_index!][error_msg]=%%C"
        set "linker_line=%%~C"
        call :tokenizeErrorLine !errwarn_index!
    )
)
exit /b

::------------------------------------------------------------------------------
:: Splits the error line from the compiler log into the filename, line number, 
:: error type, and error code for later use
::
:: Arguments: None
:: Returns:   None
::------------------------------------------------------------------------------
:tokenizeErrorLine
set "end_part="
for %%B in (error warning) do (
    if not "!error_line:*%%B=!"=="!error_line!" (
        set "end_part=%%B!error_line:*%%B=!"
    )
)

if not defined end_part (
    for /f "tokens=1,* delims=:" %%D in ("!linker_line!") do (
        set "errwarn[%~1][filename]=!error_line!"
        set "errwarn[!errwarn_index!][line_num]=%%~D"
        set "errwarn[!errwarn_index!][type]=linker"
        set "errwarn[!errwarn_index!][error_msg]=%%~E"
        set "errwarn[!errwarn_index!][error_code]=N/A"
        exit /b
    )
)

for /f "delims=" %%B in ("!end_part!") do set "front_part=!error_line:%%~B=!"
for /f "delims=" %%B in ("!front_part!") do (
	set "ext=%%~xB"
	if "!ext:~-1!"==")" (
		for /f "tokens=1,2 delims=(" %%C in ("!ext!") do (
			set "errwarn[%~1][filename]=%%~dpnB%%~C"
			set "errwarn[%~1][line_num]=%%~D"
			set "errwarn[%~1][line_num]=!errwarn[%~1][line_num]:~0,-1!"
		)
	) else (
        set "errwarn[%~1][filename]=!front_part!"
        set "errwarn[%~1][type]=linker"
    )
)
for /f "tokens=1,2" %%B in ("!end_part!") do (
	set "errwarn[%~1][type]=%%B"
	set "errwarn[%~1][error_code]=%%C"
)
exit /b

::------------------------------------------------------------------------------
:: Writes the rest of the HTML report
::
:: Arguments: None
:: Returns:   None
::------------------------------------------------------------------------------
:completeHTMLReport
>>"!logfile_header!" (
    REM == MAINTENANCE ==
    for /f "usebackq delims=" %%A in ("!logfile_maintenance!") do (
        call :alternator
        echo ^<hr^>
        echo ^<span^>%%A^</span^>
        echo ^<hr^>
        echo ^</td^>^</tr^>
    )

    REM == OPTIONS ==
    for /L %%A in (1,1,!errwarn_index!) do (
        call :alternator
        if "!errwarn[%%A][line_num]!"=="linker" (
            echo ^<hr^>
            echo ^<span style="color: red;"^>Linker Message %%A, Critical Error ^(!errwarn[%%A][error_code]!^)^</span^>
            echo ^<hr^>
            for %%B in ("!errwarn[%%A][filename]!") do echo in "%%~nxB"
            echo ^<hr^>
            echo ^<code^>!errwarn[%%A][error_msg]!^</code^>
            echo ^<hr^>
            echo ^</td^>^</tr^>
        ) else (
            echo ^<hr^>
            echo ^<span style="color: red;"^>Compiler Message %%A, !errwarn[%%A][type]! code !errwarn[%%A][error_code]!^</span^>
            for %%B in ("!errwarn[%%A][filename]!") do echo in "%%~nxB"
            echo ^<span style="color: red;"^>on line !errwarn[%%A][line_num]!^</span^>
            echo ^<hr^>
            echo ^<code^>!errwarn[%%A][error_msg]!^</code^>
            echo ^<hr^>
            echo ^</td^>^</tr^>
        )
    )

    call :alternator
    echo ^<hr^>
    if "!errwarn_index!"=="0" (
        if exist "!full_exe_path!" (
            echo ^<span^>Successfully Compiled: "!full_exe_path!"^</span^>
        ) else (
            echo ^<span style="color: red;"^>Failed Compiling: "!full_exe_path!"^</span^>
        )
    ) else (
        echo ^<span style="color: red;"^>!err_count! Errors, !warn_count! Warnings^</span^>
        echo ^<span^>from "!short_exe_path!"^</span^>
        echo ^<hr^>
        echo ^<span^>While Compiling: "!full_exe_path!"^</span^>
    )
    echo ^<hr^>
    echo ^</td^>^</tr^>

    REM == MESSAGES ==
    echo ^</table^>
    call :endHTMLReport
)

copy "!logfile_header!" "!logfile_compile!" >nul
exit /b

::------------------------------------------------------------------------------
:: Toggles the style of the table row for zebrastriping purposes
::
:: Arguments: None
:: Returns:   None
::------------------------------------------------------------------------------
:alternator
set /a OddEven=1-OddEven
echo ^<tr^>^<td class="e!OddEven!"^>
exit /b

::------------------------------------------------------------------------------
:: Kills any instances of !project_basename!.exe and any child processes
::
:: Arguments: None
:: Returns:   None
::------------------------------------------------------------------------------
:killProcess
tasklist /fi "imagename eq !short_exe_path!" 2>nul | find "!short_exe_path!" >nul
if "!errorlevel!"=="0" (
    taskkill /f /im "!short_exe_path!" /t >nul 2>&1
    call :alternator
    echo Process "!short_exe_path!" and child processes killed.
    echo ^</td^>^</tr^>
)
exit /b

::------------------------------------------------------------------------------
:: Gracefully closes the HTML report because the .exe file cannot be compiled.
:: Originally we could just abort the script entirely but we want the HTML to
:: pass a validator.
::
:: Arguments: None
:: Returns:   None
::------------------------------------------------------------------------------
:endHTMLReport
echo ^<script^>
type "!output_root!\tree.js"
type "!output_root!\clipboard.js"
echo ^</script^>
echo ^</body^>^</html^>
exit /b

::------------------------------------------------------------------------------
:: Displays the usage text for the script
::
:: Arguments: None
:: Returns:   None
::------------------------------------------------------------------------------
:usage
echo USAGE: %~nx0 ^<BASIC project folder^>
exit /b